;;;; synch.scm
;;;; Kon Lovett, Mar '06

;;; Synonyms

(define-macro (set-object!/synch ?mutex ?obj)
	`(mutex-specific-set! ,?mutex ,?obj) )

;;; Protected

(define-macro (synch MTX . BODY)
	(let ([MTX-VAR (gensym 'mtx)])
		`(let ([,MTX-VAR ,MTX])
			 (dynamic-wind
				 (lambda () (mutex-lock! ,MTX-VAR))
				 (lambda () ,@BODY)
				 (lambda () (mutex-unlock! ,MTX-VAR)))) ) )

(define-macro (synch-with MTX VAR . BODY)
	(let ([MTX-VAR (gensym 'mtx)])
		`(let ([,MTX-VAR ,MTX])
			 (let ([,VAR (mutex-specific ,MTX-VAR)])
				 (dynamic-wind
					 (lambda () (mutex-lock! ,MTX-VAR))
					 (lambda () ,@BODY)
					 (lambda () (mutex-unlock! ,MTX-VAR))))) ) )

(define-macro (call/synch MTX PROC . ARGS)
	(let ([MTX-VAR (gensym 'mtx)])
		`(let ([,MTX-VAR ,MTX])
			 (dynamic-wind
				 (lambda () (mutex-lock! ,MTX-VAR))
				 (lambda () (,PROC ,@ARGS))
				 (lambda () (mutex-unlock! ,MTX-VAR)))) ) )

(define-macro (call-with/synch MTX PROC . ARGS)
	(let ([MTX-VAR (gensym 'mtx)])
		`(let ([,MTX-VAR ,MTX])
			 (dynamic-wind
				 (lambda () (mutex-lock! ,MTX-VAR))
				 (lambda () (,PROC (mutex-specific ,MTX-VAR) ,@ARGS))
				 (lambda () (mutex-unlock! ,MTX-VAR)))) ) )

(define-macro (apply/synch MTX PROC . ARGS)
	(let ([MTX-VAR (gensym 'mtx)])
		`(let ([,MTX-VAR ,MTX])
			 (dynamic-wind
				 (lambda () (mutex-lock! ,MTX-VAR))
				 (lambda () (apply ,PROC ,@ARGS))
				 (lambda () (mutex-unlock! ,MTX-VAR)))) ) )

(define-macro (apply-with/synch MTX PROC . ARGS)
	(let ([MTX-VAR (gensym 'mtx)])
		`(let ([,MTX-VAR ,MTX])
			 (dynamic-wind
				 (lambda () (mutex-lock! ,MTX-VAR))
				 (lambda () (apply ,PROC (mutex-specific ,MTX-VAR) ,@ARGS))
				 (lambda () (mutex-unlock! ,MTX-VAR)))) ) )

(define-macro (let/synch BINDINGS . BODY)
	(car (let loop ([bnds BINDINGS])
         (cond [(null? bnds)
                 BODY]
               [(pair? (car bnds))
                 (let ([bnd (car bnds)])
                   (if (pair? bnd)
                     `((synch-with ,(cadr bnd) ,(car bnd) ,@(loop (cdr bnds))))
                     (syntax-error 'let/synch "invalid binding form" bnd) ) )]
               [else
                 (syntax-error 'let/synch "invalid binding form" bnds)] ) ) ) )

(define-macro (set!/synch ?binding . ?body)
	(if (pair? ?binding)
			(let ([?var (car ?binding)]
						[?mutex (cadr ?binding)])
				`(synch-with ,?mutex ,?var
					 (mutex-specific-set! ,?mutex (begin ,@?body))
					 (mutex-specific ,?mutex) ) )
			(syntax-error 'set!/synch "invalid binding form" ?binding) ) )

(define-macro (synch/lock MTX . BODY)
	(let ([MTX-VAR (gensym 'mtx)] [OK-VAR (gensym)] [RES-VAR (gensym)])
		`(let ([,MTX-VAR ,MTX] [,OK-VAR #f])
			 (dynamic-wind
				 (lambda () (mutex-lock! ,MTX-VAR))
				 (lambda () (let ([,RES-VAR (begin ,@BODY)]) (set! ,OK-VAR #t) ,RES-VAR))
				 (lambda () (unless ,OK-VAR (mutex-unlock! ,MTX-VAR))))) ) )

(define-macro (synch/unlock MTX . BODY)
	(let ([MTX-VAR (gensym 'mtx)])
		`(let ([,MTX-VAR ,MTX])
			 (dynamic-wind
				 (lambda ()
					 (unless (thread? (mutex-state ,MTX-VAR))
						 (warning 'synch/unlock "mutex is not locked")
						 (mutex-lock! ,MTX-VAR)))
				 (lambda () ,@BODY)
				 (lambda () (mutex-unlock! ,MTX-VAR)))) ) )

(define-macro (object/synch MTX . BODY)
	(let ([VAR (gensym)])
		(let body-loop ([unparsed BODY] [PARSED '()])
			(cond [(null? unparsed)
							`(synch-with ,MTX ,VAR ,@(reverse PARSED))]
						[(pair? unparsed)
							(let ([expr (car unparsed)]
										[next (cdr unparsed)])
								(let expr-loop ([rest expr] [EXPR '()])
									(cond [(null? rest)
													(body-loop next (cons (reverse EXPR) PARSED))]
												[(pair? rest)
													(let ([arg (car rest)]
																[next (cdr rest)])
														(if (eq? '>< arg)
																(expr-loop next (cons VAR EXPR))
																(expr-loop next (cons arg EXPR)) ) )]
												[(eq? '>< rest)
													(body-loop next (cons VAR PARSED))]
												[else
													(body-loop next (cons rest PARSED))] ) ) )]
						[else
							(syntax-error 'object/synch "invalid form" BODY)] ) ) ) )

(define-macro (record/synch SYM REC . BODY)
	`(synch (,(string->symbol (conc SYM #\- 'mutex)) ,REC)
		,@BODY) )

(define-macro (record-synch/lock SYM REC . BODY)
	`(synch/lock (,(string->symbol (conc SYM #\- 'mutex)) ,REC)
		,@BODY) )

(define-macro (record-synch/unlock SYM REC . BODY)
	`(synch/unlock (,(string->symbol (conc SYM #\- 'mutex)) ,REC)
		,@BODY) )

;;; Unprotected

(define-macro (%synch-mutex* MTX . BODY)
	(let ([RET-VAR (gensym)]
	      [MTX-VAR (gensym 'mtx)])
		`(let ([,MTX-VAR ,MTX])
			 (begin
				 (mutex-lock! ,MTX-VAR)
				 (call-with-values
					 (lambda () ,@BODY)
					 (lambda ,RET-VAR
						 (mutex-unlock! ,MTX-VAR)
						 (apply values ,RET-VAR))))) ) )

(define-macro (%synch-mutex-with* MTX VAR . BODY)
	(let ([RET-VAR (gensym)]
	      [MTX-VAR (gensym 'mtx)])
		`(let ([,MTX-VAR ,MTX])
			 (let ([,VAR (mutex-specific ,MTX-VAR)])
				 (mutex-lock! ,MTX-VAR)
				 (call-with-values
					 (lambda () ,@BODY)
					 (lambda ,RET-VAR
						 (mutex-unlock! ,MTX-VAR)
						 (apply values ,RET-VAR))))) ) )

(define-macro (%synch MTX . BODY)
	`(%synch-mutex* ,MTX ,@BODY) )

(define-macro (%synch-with MTX VAR . BODY)
	`(%synch-mutex-with* ,MTX ,VAR ,@BODY) )

(define-macro (%call/synch MTX PROC . ARGS)
	`(%synch-mutex* ,MTX (,PROC ,@ARGS)) )

(define-macro (%call-with/synch MTX PROC . ARGS)
	(let ([VAR (gensym)])
		`(%synch-mutex-with* ,MTX ,VAR (,PROC ,VAR ,@ARGS)) ) )

(define-macro (%apply/synch MTX PROC . ARGS)
	`(%synch-mutex* ,MTX (apply ,PROC ,@ARGS)) )

(define-macro (%apply-with/synch MTX PROC . ARGS)
	(let ([VAR (gensym)])
		`(%synch-mutex-with* ,MTX ,VAR (apply ,PROC ,VAR ,@ARGS)) ) )

(define-macro (%let/synch BINDINGS . BODY)
	(car (let loop ([bnds BINDINGS])
         (cond [(null? bnds)
                 BODY]
               [(pair? (car bnds))
                 (let ([bnd (car bnds)])
                   (if (pair? bnd)
                       `((%synch-with ,(cadr bnd) ,(car bnd) ,@(loop (cdr bnds))))
                       (syntax-error '%let/synch "invalid binding form" bnd) ) )]
               [else
                 (syntax-error '%let/synch "invalid binding form" bnds)] ) )) )

(define-macro (%set!/synch ?binding . ?body)
	(if (pair? ?binding)
			(let ([?var (car ?binding)]
						[?mutex (cadr ?binding)])
				`(%synch-with ,?mutex ,?var
					 (mutex-specific-set! ,?mutex (begin ,@?body))
					 (mutex-specific ,?mutex) ) )
			(syntax-error '%set!/synch "invalid binding form" ?binding) ) )

(define-macro (%synch/lock MTX . BODY)
	(let ([RET-VAR (gensym)] [MTX-VAR (gensym 'mtx)] [OK-VAR (gensym)] [RES-VAR (gensym)])
		`(let ([,MTX-VAR ,MTX] [,OK-VAR #f])
			(begin
				(mutex-lock! ,MTX-VAR)
				(call-with-values
					(lambda () (let ([,RES-VAR (begin ,@BODY)]) (set! ,OK-VAR #t) ,RES-VAR))
					(lambda ,RET-VAR
						(unless ,OK-VAR (mutex-unlock! ,MTX-VAR))
						(apply values ,RET-VAR))) ) ) ) )

(define-macro (%synch/unlock MTX . BODY)
	(let ([RET-VAR (gensym)] [MTX-VAR (gensym 'mtx)])
		`(let ([,MTX-VAR ,MTX])
			 (begin
				 (unless (thread? (mutex-state ,MTX-VAR))
					 (warning '%synch/unlock "mutex is not locked")
					 (mutex-lock! ,MTX-VAR))
				 (call-with-values
					 (lambda () ,@BODY)
					 (lambda ,RET-VAR
						 (mutex-unlock! ,MTX-VAR)
						 (apply values ,RET-VAR)) ) ) ) ) )

(define-macro (%object/synch MTX . BODY)
	(let ([VAR (gensym)])
		(let body-loop ([unparsed BODY] [PARSED '()])
			(cond [(null? unparsed)
							`(%synch-with ,MTX ,VAR ,@(reverse PARSED))]
						[(pair? unparsed)
							(let ([expr (car unparsed)]
										[next (cdr unparsed)])
								(let expr-loop ([rest expr] [EXPR '()])
									(cond [(null? rest)
													(body-loop next (cons (reverse EXPR) PARSED))]
												[(pair? rest)
													(let ([arg (car rest)]
																[next (cdr rest)])
														(if (eq? '>< arg)
																(expr-loop next (cons VAR EXPR))
																(expr-loop next (cons arg EXPR)) ) )]
												[(eq? '>< rest)
													(body-loop next (cons VAR PARSED))]
												[else
													(body-loop next (cons rest PARSED))] ) ) )]
						[else
							(syntax-error 'object/synch "invalid form" BODY)] ) ) ) )

(define-macro (%record/synch SYM REC . BODY)
	`(%synch (,(string->symbol (conc SYM #\- 'mutex)) ,REC)
		,@BODY) )

(define-macro (%record-synch/lock SYM REC . BODY)
	`(%synch/lock (,(string->symbol (conc SYM #\- 'mutex)) ,REC)
		,@BODY) )

(define-macro (%record-synch/unlock SYM REC . BODY)
	`(%synch/unlock (,(string->symbol (conc SYM #\- 'mutex)) ,REC)
		,@BODY) )

;;; DEPRECATED

(define-macro (set-synchronized-object! ?mutex ?obj)
	`(mutex-specific-set! ,?mutex ,?obj) )

(define-macro (synch/obj MTX . BODY)
	(let ([VAR (gensym)])
		(let body-loop ([unparsed BODY] [PARSED '()])
			(cond [(null? unparsed)
							`(synch-with ,MTX ,VAR ,@(reverse PARSED))]
						[(pair? unparsed)
							(let ([expr (car unparsed)]
										[next (cdr unparsed)])
								(let expr-loop ([rest expr] [EXPR '()])
									(cond [(null? rest)
													(body-loop next (cons (reverse EXPR) PARSED))]
												[(pair? rest)
													(let ([arg (car rest)]
																[next (cdr rest)])
														(if (eq? '>< arg)
																(expr-loop next (cons VAR EXPR))
																(expr-loop next (cons arg EXPR)) ) )]
												[(eq? '>< rest)
													(body-loop next (cons VAR PARSED))]
												[else
													(body-loop next (cons rest PARSED))] ) ) )]
						[else
							(syntax-error 'synch/obj "invalid form" BODY)] ) ) ) )

(define-macro (synch/record SYM REC . BODY)
	`(synch (,(string->symbol (conc SYM #\- 'mutex)) ,REC)
		,@BODY) )

(define-macro (synch/lock-record SYM REC . BODY)
	`(synch/lock (,(string->symbol (conc SYM #\- 'mutex)) ,REC)
		,@BODY) )

(define-macro (synch/unlock-record SYM REC . BODY)
	`(synch/unlock (,(string->symbol (conc SYM #\- 'mutex)) ,REC)
		,@BODY) )

(define-macro (%synch/obj MTX . BODY)
	(let ([VAR (gensym)])
		(let body-loop ([unparsed BODY] [PARSED '()])
			(cond [(null? unparsed)
							`(%synch-with ,MTX ,VAR ,@(reverse PARSED))]
						[(pair? unparsed)
							(let ([expr (car unparsed)]
										[next (cdr unparsed)])
								(let expr-loop ([rest expr] [EXPR '()])
									(cond [(null? rest)
													(body-loop next (cons (reverse EXPR) PARSED))]
												[(pair? rest)
													(let ([arg (car rest)]
																[next (cdr rest)])
														(if (eq? '>< arg)
																(expr-loop next (cons VAR EXPR))
																(expr-loop next (cons arg EXPR)) ) )]
												[(eq? '>< rest)
													(body-loop next (cons VAR PARSED))]
												[else
													(body-loop next (cons rest PARSED))] ) ) )]
						[else
							(syntax-error 'synch/obj "invalid form" BODY)] ) ) ) )

(define-macro (%synch/record SYM REC . BODY)
	`(%synch (,(string->symbol (conc SYM #\- 'mutex)) ,REC)
		,@BODY) )

(define-macro (%synch/lock-record SYM REC . BODY)
	`(%synch/lock (,(string->symbol (conc SYM #\- 'mutex)) ,REC)
		,@BODY) )

(define-macro (%synch/unlock-record SYM REC . BODY)
	`(%synch/unlock (,(string->symbol (conc SYM #\- 'mutex)) ,REC)
		,@BODY) )
